#include <Servo.h>
#include <SharpDistSensor.h>
#include <AccelStepper.h>
#include <MultiStepper.h>

//determined by experimentation.  
//Will need to be updated if robot dimensions don't exactly match instructable
float DEGREES_PER_STEP = 66.8/800; 

//########################################################
//###############DECLARATIONS#############################
//########################################################
///////////////////SERVO/////////////////////////////

Servo Neck;  // Rangefinder Positioner

Servo Antenna1;  //Mood Display
Servo Antenna2;  //Mood Display

int moodPos1 = 90; //servo position
int moodPos2 = 90; //servo position
int moodTime; //for mood states that move the "antennae"
int moodDir; //for oscillating mood states
int Mood = 0; // the current mood

int pos = 90;    // the neck position
long MoveTime = 0; //for controlling neck speed
int ServoSpeedDelay = 10; //for controlling neck speed
bool RotateDir = true; //for alternating turn direction

///////////////////RANGEFINDER/////////////////////////////

// Analog pin to which the sensor is connected
const byte sensorPin = A0;

// Window size of the median filter (odd number, 1 = no filtering)
const byte medianFilterWindowSize = 5;

// Create an object instance of the SharpDistSensor class
SharpDistSensor RangeFinder(sensorPin, medianFilterWindowSize);

unsigned int Range = 1500;  //the current range reading

int PreviousRange = 1500; //for tracking acquired objects
int AcquireRange = 600; //the threshold for "noticing" an object
int FindRange = 150; //the threshold for "arriving" at an object
int ClosestRange = 1500; //for tracking object (re)acquisition
int ClosestPos = 90; //for tracking object (re)acquisition

///////////////////KICKER/////////////////////////////
int EndStopPin_Gripped = 28;
int EndStopPin_Loaded = 27;
int EndStopPin_Fired = 29;
int HBridgePin_IN1 = 11;
int HBridgePin_IN2 = 10;
int HBridgePin_enA = 8;

int KickerCount = 0; //for timing the reload motion
int KickerStopCount = 40; //for timing the reload motion
int KickerSpeed = 255;  //kicker motor speed
bool Grabbed = false;  //tracks if the bolt head is gripped
bool Loaded = false;  //tracks if the kicker is ready to fire
int EndStop_Gripped;  //end stop value
int EndStop_Fired;  //end stop value
///////////////////STEPPERS/////////////////////////////

AccelStepper RightFrontWheel(1, 40, 41); // Stepper1
AccelStepper RightBackWheel(1, 20, 21);   // (Type:driver, STEP, DIR) - Stepper2
AccelStepper LeftFrontWheel(1, 44, 45);  // Stepper3
AccelStepper LeftBackWheel(1, 16, 17);  // Stepper4

MultiStepper steppers;

long positions[4] = {0,0,0,0};
int wheelSpeed = 1000;
int StepIncrement = 1;

////////////////EDGE SENSORS//////////////////////////////
int S_rr = 22;
int S_lr = 38;
int S_rf = 23;
int S_lf = 43;

int Sensor_rr;
int Sensor_lr;
int Sensor_rf;
int Sensor_lf;

int Sensor_rr_stored = 0;  //for tracking previous value
int Sensor_lr_stored = 0;  //for tracking previous value
int Sensor_rf_stored = 0;  //for tracking previous value
int Sensor_lf_stored = 0;  //for tracking previous value

////////////////LOGIC//////////////////////////////
bool ObjectDetected = false;  //tracks if the bot has seen something
bool ObjectLost = false;  //tracks if the bot has lost track of an acquired object
bool ObjectFound = false;  //tracks if the bot has arrived at an object
bool NewCommandNeeded = true;  //tracks if the bot should rethink it's direction in life
bool EdgeDetected = false;  //tracks edge detection

bool Reacquire = false;  //turns off neck swivel during reacquire move
bool SweepStart = false;  //tracks reacquire motion
bool SweepEnd = false; //tracks reacquire motion
bool BlindMove = false;  //turns off object acquisition for one move duration

int KickSelection = 0;  // tracks selection of what to do when an object is reached - kick, nudge, ignore etc.
int IdlePick = 0;  //tracks what to do when no object is spotted.  Sit, wander, etc.
int IdleStep = 0;  //tracks multi-step idle moves, e.g. rotate then move forwards

long Clock = millis();  //tracks time

float MovementCommand [4] = {0,0,0,0};  //tracks movement direciton

int Counter = 0;  //tracks movement duration (in steps)
int CounterLimit = 1000;  //tracks when a move is finished


///////////////////MOVE DIRECTIONS//////////////////////////
//These are currently not used in the code!
//they are mostly here to help you understand which direction the steppers should turn for a corresponding move
float Stop [4] =  {0,0,0,0}; //Stop
float Forward [4] =   {-1,-1,1,1}; //Forward
float Backward [4] =  {1,1,-1,-1}; //Backward
float Right [4] =  {-1,1,-1,1}; //Right
float Left [4] =  {1,-1,1,-1}; //Left
float ForwardRight [4] =  {-1,0,0,1}; //Forward Right
float BackwardLeft [4] =  {1,0,0,-1}; //Backward Left
float ForwardLeft [4] =  {0,-1,1,0}; //Forward Left
float BackwardRight [4] =  {0,1,-1,0}; //Backward Right


//########################################################
//###########################SETUP########################
//########################################################
void setup() 
{

  pinMode(EndStopPin_Gripped, INPUT_PULLUP);
  pinMode(EndStopPin_Loaded, INPUT);
  pinMode(EndStopPin_Fired, INPUT);
  pinMode(HBridgePin_IN1, OUTPUT);
  pinMode(HBridgePin_IN2, OUTPUT);
  pinMode(HBridgePin_enA, OUTPUT);
  
  pinMode(S_rf,INPUT);
  pinMode(S_lf,INPUT);
  pinMode(S_rr,INPUT);
  pinMode(S_lr,INPUT);
    
  LeftFrontWheel.setMaxSpeed(200);
  RightFrontWheel.setMaxSpeed(200);
  RightBackWheel.setMaxSpeed(200);
  LeftBackWheel.setMaxSpeed(200);

  steppers.addStepper(LeftFrontWheel);
  steppers.addStepper(LeftBackWheel);
  steppers.addStepper(RightFrontWheel);
  steppers.addStepper(RightBackWheel);


  //servos
  Neck.attach(4);  
  Antenna1.attach(2);
  Antenna2.attach(13);
  
  
  //Serial.begin(9600);  //You can uncomment serial lines below for debugging or testing

}

//########################################################
//###########################LOOP#########################
//########################################################
void loop() 
{
  //Recompute where to drive
  if (NewCommandNeeded)
  {
    BlindMove = false;  //finish any blind move
    MotionControl();  //pick a new command
    NewCommandNeeded = false;
    Counter = 0;  //reset the move counter
  }
  
  //Keep going until you need to change directions
  while (not NewCommandNeeded)
  {
    
    //check sensors
    InterpretEdges();  //check for edge detection
    InterpretRange();  //check for object spotting
    //InterpretSound();
  
    //constantly reload kicker
    if (not Loaded){
      CheckKicker();
    }
  
    //keep driving
    executeMotions(MovementCommand);
    Counter ++;
    UpdateMood();  //keep expressing yourself
    if (Counter >= CounterLimit)
    {
      NewCommandNeeded = true;
      //if the move is finished, ask for a new move
    }
  }
}

//########################################################
//#######################FUNCTIONS########################
//########################################################

///////////////////////MOTION////////////////////////////


//this function is the "brain" that decides how to move when idle or reacting to stimulus
void MotionControl()
{
  if (EdgeDetected) //this overrides all other possibilities.  The robot should always avoid falling!
  {
    Mood = 2;
//    Serial.print("edge detected! ");
//    Serial.print(Sensor_rf);
//    Serial.print(Sensor_rr);
//    Serial.print(Sensor_lf);
//    Serial.println(Sensor_lr);
    EdgeCorrect();
  }
  else if (ObjectDetected)  //this variable is set when the sensor is interpreted
  {
    if (ObjectFound)
    {
      //Serial.println("FIRE!!!!!!!!");
      FindObject();  //arrive at an object
    }
    else
    {
      if (ObjectLost)  //this variable is set when the sensor is interpreted
      {
//        Serial.print("reacquiring.....");
//        Serial.print(PreviousRange);
//        Serial.print(" , ");
//        Serial.println(Range);
        ReacquireObject();
      }
      else
      {
        //keep moving towards object
        MovementCommand[0] = -1; // forwards
        MovementCommand[1] = -1; // forwards
        MovementCommand[2] = 1; // forwards
        MovementCommand[3] = 1; // forwards
      }
    }
  }
  else
  {
    if (IdlePick == 0)
    {
      //if the robot is idle, pick a new behavoir
      IdlePick = random(1,2);
    }
    switch(IdlePick)
    //BE CREATIVE HERE#######################################################################
    //You can easily write new idle behavoirs, and simply extend the range of the random value above
    {
      case 1:  //wait
        //Serial.println("waiting....");
        MovementCommand[0] = 0; // stop
        MovementCommand[1] = -0; // stop
        MovementCommand[2] = 0; // stop
        MovementCommand[3] = 0; // stop
        CounterLimit = random(1000,10000);
        Mood = 0;
        IdlePick = 0;
        break;
      case 2: //new direction
        if (IdleStep == 0)
        {
          //Serial.println("reorienting");
          Rotate(random(-180,180));
          IdleStep = 1;
        }
        else if (IdleStep == 1)
        {
          MovementCommand[0] = -1; // forwards
          MovementCommand[1] = -1; // forwards
          MovementCommand[2] = 1; // forwards
          MovementCommand[3] = 1; // forwards
          CounterLimit = random(500,2000);
          Mood = 3;
          //Serial.println("moving");
          IdleStep = 0;
          IdlePick = 0;
        }
        break;
      case 3: //90 turn
        Rotate(90);
        
    }
  }
}

void EdgeCorrect() 
//move away from a detected edge
{
  CounterLimit = 500;  //how long to move away for
  Counter = 0;
  ObjectDetected = false;  //forget any spotted objects (they could be off the table)
  ResetObject(); //forget any spotted objects (they could be off the table)
  if (Sensor_rf == 1 && Sensor_lf == 1) 
  {
    MovementCommand[0] = 1; // backwards
    MovementCommand[1] = 1; // backwards
    MovementCommand[2] = -1; // backwards
    MovementCommand[3] = -1; // backwards
  }
  else if (Sensor_rf == 1 && Sensor_rr == 1) 
  {
    MovementCommand[0] = -1; // right
    MovementCommand[1] = 1; // right
    MovementCommand[2] = -1; // right
    MovementCommand[3] = 1; // right
  }
  else if (Sensor_rr == 1 && Sensor_lr == 1)
  {
    MovementCommand[0] = -1; // forwards
    MovementCommand[1] = -1; // forwards
    MovementCommand[2] = 1; // forwards
    MovementCommand[3] = 1; // forwards
  }
  else if (Sensor_lr == 1 && Sensor_lf == 1)
  {
    MovementCommand[0] = 1; // left
    MovementCommand[1] = -1; // left
    MovementCommand[2] = 1; // left
    MovementCommand[3] = -1; // left
  }
  else if (Sensor_rf == 1)
  {
    MovementCommand[0] = 1; // backwards left
    MovementCommand[1] = 0; // backwards left
    MovementCommand[2] = 0; // backwards left
    MovementCommand[3] = -1; // backwards left
  }
  else if (Sensor_lf == 1)
  {
    MovementCommand[0] = 0; // backwards right
    MovementCommand[1] = 1; // backwards right
    MovementCommand[2] = -1; // backwards right
    MovementCommand[3] = 0; // backwards right
  }
  else if (Sensor_rr == 1)
  {
    MovementCommand[0] = 0; // forwards left
    MovementCommand[1] = -1; // forwards left
    MovementCommand[2] = 1; // forwards left
    MovementCommand[3] = 0; // forwards left
  }
  else if (Sensor_lr == 1)
  {
    MovementCommand[0] = -1; // forwards right
    MovementCommand[1] = 0; // forwards right
    MovementCommand[2] = 0; // forwards right
    MovementCommand[3] = 1; // forwards right
  }
}

void AcquireObject()
{
  int Target = -(ClosestPos - 90);  //the angle to rotate to
  //Serial.print("target acquired at ");
  //Serial.println(Target);
  ObjectDetected = true;
  PreviousRange = ClosestRange;
  RotateFront(Target);
  BlindMove = true;
  Neck.write(90);
  pos = 90;
  Mood = 1;
  UpdateMood();
  delay(50);
}

void FindObject()  //arrive at an ojbect
{
  if (KickSelection == 0)  // randomly pick an encounter behavior
  {
    executeMotions(Stop);
    delay(1000);
    KickSelection = random(1,3);
  }
  switch(KickSelection)
  //BE CREATIVE HERE#######################################################################
  //You can easily write new encounter behavoirs (ignore, etc.) and extend the range
  //of the random value above
  {
    case 1: //nudge
      if (ObjectDetected)
      {
        delay(random(200,1000));
        MovementCommand[0] = -1; // forwards
        MovementCommand[1] = -1; // forwards
        MovementCommand[2] = 1; // forwards
        MovementCommand[3] = 1; // forwards
        Counter = 0;
        CounterLimit = random(50,500);
        //Serial.println(CounterLimit);
      }
      else
      {
        ResetObject();
      }
      break;
    case 2:  //kick
      if (ObjectDetected) 
      {
        float nudge [4] =  {-50,-50,50,50};
        executeMotions(nudge);
        delay(1000);
        Kick();
        ObjectDetected = false;
      }
      else
      {
        ResetObject();
      }
      break;
  }
}

void ReacquireObject()  //try to find an object again if the robot has lost track
{
  Reacquire = true;
  if (SweepStart && SweepEnd)
  //if the sweep has completed without reacquring the object, forget the object
  {
    ObjectDetected = false;
    ResetObject();
    //Serial.println("target lost");
    Reacquire = false;
    ObjectLost = false;
    SweepStart = false;
    SweepEnd = false;
    RotateDir = true; 
    }
  else if (SweepStart)
  {
    if (Range < PreviousRange + 50)
    //find the ojbect
    {
      ClosestRange = Range;
      ClosestPos = pos;
      //Serial.print("target reacquired at ");
      //Serial.println(ClosestPos); 
      Reacquire = false;  
      ObjectLost = false;
      SweepStart = false;
      SweepEnd = false;
      RotateDir = true; 
    }
    else
    //continue the sweep
    {
      Rotate(1);
    //Serial.print("sweep running...");
    //Serial.print(pos);
    //Serial.print(" , ");
    //Serial.println(Range);
    pos++;
    }
    if (pos > 60)
    {
      SweepEnd = true;
      //Serial.println("sweep finished");
    }
  }
  else
  //start the sweep
  {
    Rotate(-30);
    SweepStart = true;
    pos = 0;
    ClosestRange = 1500;
    ClosestPos = 0;
    //Serial.println("sweep starting...");
  }
}

void ResetObject()
//forget any object and rest all object-related parameters
{
  KickSelection = 0;
  ClosestRange = 1500;
  ClosestPos = 90;
  NewCommandNeeded = true;
  ObjectLost = false;
  ObjectFound = false;
  Mood = 0;
  Reacquire = false;
  SweepStart = false;
  SweepEnd = false;
  RotateDir = true;
}

void Rotate(float TargetAngle)
//calculate how far to move for a given angular rotation
{
  Counter = 0;
  float TargetSteps = TargetAngle/DEGREES_PER_STEP;
  float StepCount = TargetSteps/StepIncrement;
  //Serial.print("rotate ");
  //Serial.print(TargetAngle);
  //Serial.print(" , ");
  if (StepCount < 0)
  {
    StepCount = -StepCount;
  }
  //Serial.println(StepCount);
  if (TargetAngle > 0)
  {
    MovementCommand[0] = 1;
    MovementCommand[1] = 1;
    MovementCommand[2] = 1;
    MovementCommand[3] = 1;
  }
  else if (TargetAngle < 0)
  {
    MovementCommand[0] = -1;
    MovementCommand[1] = -1;
    MovementCommand[2] = -1;
    MovementCommand[3] = -1;
  }
  CounterLimit = StepCount;
}

void RotateFront(float TargetAngle) //Rotate around the front pivot point
{
  Counter = 0;
  float TargetSteps = TargetAngle/DEGREES_PER_STEP;
  float StepCount = TargetSteps/StepIncrement;
  //Serial.print("rotate front ");
  //Serial.print(TargetAngle);
  //Serial.print(" , ");
  if (StepCount < 0)
  {
    StepCount = -StepCount;
  }
  //Serial.println(StepCount);
  if (TargetAngle > 0)
  {
    MovementCommand[0] = 0;
    MovementCommand[1] = 1;
    MovementCommand[2] = 0;
    MovementCommand[3] = 1;
  }
  else if (TargetAngle < 0)
  {
    MovementCommand[0] = 0;
    MovementCommand[1] = -1;
    MovementCommand[2] = 0;
    MovementCommand[3] = -1;
  }
  CounterLimit = StepCount;
}

void executeMotions(float Command[4]) 
//keep stepping
{
  positions[0] += StepIncrement * Command[0];
  positions[1] += StepIncrement * Command[1];
  positions[2] += StepIncrement * Command[2];
  positions[3] += StepIncrement * Command[3];
  steppers.moveTo(positions);
  steppers.runSpeedToPosition();
}

///////////////////////SENSORS////////////////////////////

void InterpretRange()
//read the range finder and decide if an object has been found, lost etc.
{
  //Serial.println(Range);
  if (BlindMove)
  {
    //don't look
  }
  else if (Reacquire)
  {
    //look forwards
    Neck.write(90);
    Range = RangeFinder.getDist();
  }
  else
  {
    Range = RangeFinder.getDist();
    if (ObjectDetected)
    {
      if (Range > PreviousRange + 50)
      //lose track
      {
        if (not ObjectFound)
        {
          ObjectLost = true;
          NewCommandNeeded = true;
        }
      }
      else if (Range < FindRange)
      //arrive at object
      {
        if (not ObjectFound)
        {
          ObjectFound = true;
          NewCommandNeeded = true;
        }
      }
      else
      {
        //keep going
        PreviousRange = Range;
      }
    }
    else if (Range < AcquireRange)
    {
      if (Range < ClosestRange)
      {
        //notice a potental object.
        //if the robot is stationary, it will complete the neck swivel before acquiring
        ClosestRange = Range;
        ClosestPos = pos;
      }
      if (MovementCommand[0] != 0 || MovementCommand[1] != 0 || MovementCommand[2] != 0 || MovementCommand[3] != 0)
      {
        AcquireObject();
      }
      else
      {
        UpdateServo();
      }
    }
    else
    {
      UpdateServo();
      //continue swiveling
    }
  }
}

void InterpretEdges()
//check if an edge has been (un-)spotted
{
  Sensor_rf = digitalRead(S_rf);
  Sensor_lf = digitalRead(S_lf);
  Sensor_rr = digitalRead(S_rr);
  Sensor_lr = digitalRead(S_lr);
  if (not EdgeDetected)
  {
    if (Sensor_rf == 1 || Sensor_lf == 1 || Sensor_rr == 1 || Sensor_lr == 1) {
      NewCommandNeeded = true;
      EdgeDetected = true;
      ObjectDetected = false;
      ResetObject();
      Sensor_rr_stored = Sensor_rr;
      Sensor_lr_stored = Sensor_lr;
      Sensor_rf_stored = Sensor_rf;
      Sensor_lf_stored = Sensor_lf;
    }
  }
  else
  {
    if (Sensor_rf == 1 && Sensor_rf != Sensor_rf_stored)
    {
      Sensor_rf_stored = 1;
      NewCommandNeeded = true;
    }
    else if (Sensor_lf == 1 && Sensor_lf != Sensor_lf_stored)
    {
      Sensor_lf_stored = 1;
      NewCommandNeeded = true;
    }
    else if (Sensor_rr == 1 && Sensor_rr != Sensor_rr_stored)
    {
      Sensor_rr_stored = 1;
      NewCommandNeeded = true;
    }
    else if (Sensor_lr == 1 && Sensor_lr != Sensor_lr_stored)
    {
      Sensor_lr_stored = 1;
      NewCommandNeeded = true;
    }
    else if (Sensor_rf == 0 && Sensor_lf == 0 && Sensor_rr == 0 && Sensor_lr == 0)
    {
      EdgeDetected = false;
      Sensor_rr_stored = Sensor_rr;
      Sensor_lr_stored = Sensor_lr;
      Sensor_rf_stored = Sensor_rf;
      Sensor_lf_stored = Sensor_lf;
    }
  }
}

void UpdateServo()
//swivel the neck
{
  if (millis() >= MoveTime)
  {
    if (RotateDir){
      if (pos == 150){
        RotateDir = false;
      }
      else
      {
        pos++;
        Neck.write(pos);
      }
    }
    else
    {
      if (pos == 30) //sweep complete
      {
        RotateDir = true;
        if (ClosestRange < AcquireRange)  //if something got spotted on this sweep, lock onto it
        {
          AcquireObject();
        }
      }
      else
      {
        pos--;
        Neck.write(pos);
      }
    }
    MoveTime = millis() + ServoSpeedDelay;
  }
}

///////////////////////KICKER////////////////////////////

void CheckKicker()
//keep reloading the kicker, if needed
//check end stops to see if the kicker is gripped or loaded
{
  EndStop_Gripped = digitalRead(EndStopPin_Gripped);
  if (not Grabbed)
  {
    if (EndStop_Gripped == 1)
    {
      Grabbed = true;
      KickerCount = 0;
    }
    else
    {
      analogWrite(HBridgePin_enA,KickerSpeed);
      digitalWrite(HBridgePin_IN1, LOW);
      digitalWrite(HBridgePin_IN2, HIGH);
    }
  }
  else
  {
    if (EndStop_Gripped == 0)
    {
      if (KickerCount >= KickerStopCount)
      {
        analogWrite(HBridgePin_enA,0);
        Loaded = true;
      }
      else
      {
        KickerCount++;
      }
    }
    else
    {
      analogWrite(HBridgePin_enA,KickerSpeed);
      digitalWrite(HBridgePin_IN1, HIGH);
      digitalWrite(HBridgePin_IN2, LOW);
    }
  }
}

void Kick()
//fire the kicker
//freezes all other operations until finished
{
  while (Loaded)
  {
    EndStop_Fired = digitalRead(EndStopPin_Fired);
    if (EndStop_Fired == 0)
    {
      analogWrite(HBridgePin_enA,0);
      Loaded = false;
      Grabbed = false;
    }
    else
    {
      analogWrite(HBridgePin_enA,KickerSpeed);
      digitalWrite(HBridgePin_IN1, HIGH);
      digitalWrite(HBridgePin_IN2, LOW);
    }
  }
}

void UpdateMood()
//display the current mood
{
  if (Mood == 0) // normal
  {
    moodPos1 = 90;
    moodPos2 = 90;
  }
  if (Mood == 1) //alert
  {
    moodPos1 = 180;
    moodPos2 = 0;
  }
  if (Mood == 2) //edge detected
  {
    if (millis() > moodTime)
    {
      if (moodDir)
      {
        moodPos1 = 60;
        moodPos2 = 60;
        moodDir = false;
      }
      else
      {
        moodPos1 = 120;
        moodPos2 = 120;
        moodDir = true;
      }
    }
    moodTime = millis() + 100;
  }
  if (Mood == 3) //moving
  {
    if (moodPos1 > 120)
    {
      moodDir = false;
    }
    if (moodPos1 < 60)
    {
      moodDir = true;
    }
    if (millis() >= moodTime)
    {
      if (moodDir)
      {
        if (moodPos1 == 120)
        {
          moodDir = false;
        }
        else
        {
          moodPos1++;
          moodPos2++;
        }
      }
      if (not RotateDir)
      {
        if (moodPos1 == 60)
        {
          moodDir = true;
        }
        else
        {
          moodPos1--;
          moodPos2--;
        }
      }
    }
    moodTime = millis() + 20;
  }
  Antenna1.write(moodPos1);
  Antenna2.write(moodPos2);
}
